'''dossier.fc Feature Collections

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

'''
from __future__ import absolute_import, division, print_function
try:
    from collections import Counter
except ImportError:
    from backport_collections import Counter

import pytest

from dossier.fc import FeatureCollection, StringCounter, ReadOnlyException
from dossier.fc.tests import counter_type


def test_read_only():
    fcwork = FeatureCollection({'feat': {'foo': 1}})

    fc = FeatureCollection()
    fc['feat']['foo'] += 1
    fc.read_only = True
    with pytest.raises(ReadOnlyException):
        fc += fcwork
    with pytest.raises(ReadOnlyException):
        fc -= fcwork
    with pytest.raises(ReadOnlyException):
        fc -= fcwork
    with pytest.raises(ReadOnlyException):
        del fc['feat']
    with pytest.raises(ReadOnlyException):
        fc.pop('feat')


def test_read_only_binop():
    fc1 = FeatureCollection({'NAME': {'foo': 1, 'bar': 1}})
    fc2 = FeatureCollection({'NAME': {'foo': 2, 'bar': 2}})

    fc1.read_only = True
    fc2.read_only = True

    result = fc1 + fc2
    expected = FeatureCollection({'NAME': {'foo': 3, 'bar': 3}})
    assert result == expected
    assert not result.read_only


def test_read_only_features():
    fc = FeatureCollection({'feat': StringCounter({'foo': 1})})
    fc['feat']['foo'] += 1
    fc.read_only = True

    with pytest.raises(ReadOnlyException):
        fc['feat']['foo'] += 1
    with pytest.raises(ReadOnlyException):
        fc['feat'].pop('foo')
    with pytest.raises(ReadOnlyException):
        del fc['feat']['foo']


def test_identity():
    fc = FeatureCollection()
    fc.read_only = True
    id(fc['one']) == id(fc['two'])


def test_read_only_preserved_after_serialized():
    fc = FeatureCollection({'NAME': {'foo': 1, 'baz': 2}})
    fc.read_only = True
    fcnew = FeatureCollection.loads(fc.dumps())
    assert fcnew.read_only
    with pytest.raises(ReadOnlyException):
        fcnew['NAME']['foo'] += 1


def test_read_only_not_preserved_via_dict():
    fc = FeatureCollection({'NAME': {'foo': 1, 'baz': 2}})
    fc.read_only = True
    fcnew = FeatureCollection(fc.to_dict())
    assert not fcnew.read_only
    fcnew['NAME']['foo'] += 1


def test_readonly(counter_type):
    fc = FeatureCollection({
            'hello': counter_type(Counter('hello')),
            'goodbye': counter_type(Counter('goodbye'))})
    fc2 = FeatureCollection({
            'hello': counter_type(Counter('hello')),
            'goodbye': counter_type(Counter('goodbye'))})

    fc.read_only = True
    with pytest.raises(ReadOnlyException):
        fc += fc2

    with pytest.raises(ReadOnlyException):
        fc -= fc2

    with pytest.raises(ReadOnlyException):
        fc *= 2

    with pytest.raises(ReadOnlyException):
        fc['woof'] = StringCounter()

    if hasattr(counter_type, 'read_only'):
        with pytest.raises(ReadOnlyException):
            fc['hello']['l'] = 3
        with pytest.raises(ReadOnlyException):
            fc['hello']['l'] += 3

    fc.read_only = False
    fc += fc2
    assert Counter(map(abs,fc['hello'].values())) == Counter({2: 3, 4: 1})
    fc -= fc2
    fc -= fc2
    assert Counter(map(abs,fc['hello'].values())) == Counter()



from __future__ import absolute_import, division, print_function
import abc


class SparseVector(object):
    '''An abstract class for sparse vectors.

    Currently, there is no default implementation of a sparse vector.

    Other implementations of sparse vectors *must* inherit
    from this class. Otherwise they cannot be used inside a
    :class:`dossier.fc.FeatureCollection`.
    '''
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __init__(self): pass


class DenseVector(object):
    '''An abstract class for dense vectors.

    Currently, there is no default implementation of a dense vector.

    Other implementations of dense vectors *must* inherit
    from this class. Otherwise they cannot be used inside a
    :class:`dossier.fc.FeatureCollection`.
    '''
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __init__(self): pass



'''dossier.fc Feature Collections

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

'''
from __future__ import absolute_import, division, print_function
try:
    from collections import Counter
except ImportError:
    from backport_collections import Counter
from operator import itemgetter

from dossier.fc.tests import counter_type  # noqa


def test_truncation(counter_type):  # noqa
    num = 100
    data = {}
    for x in xrange(num):
        data[str(x)] = x + 1

    counter = counter_type(data)
    assert len(counter) == num

    truncation_length = 10
    counter.truncate_most_common(truncation_length)
    assert len(counter) == truncation_length

    most_common = Counter(data).most_common(truncation_length)
    from_counter = map(itemgetter(1), counter.most_common(truncation_length))
    expected = map(itemgetter(1), most_common)
    from_counter = map(abs, from_counter)
    assert from_counter == expected

    assert set(counter_type(dict(most_common)).items()) == set(counter.items())

    expected = {}
    for x in xrange(90, 100):
        expected[str(x)] = x + 1
    should_be_counter = counter_type(expected)
    assert should_be_counter == counter



'''dossier.fc Feature Collections

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

.. autoclass:: StringCounter
'''
from __future__ import absolute_import, division, print_function
try:
    from collections import Counter
except ImportError:
    from backport_collections import Counter
from collections import Mapping, MutableMapping
from functools import wraps

from dossier.fc.exceptions import ReadOnlyException, uni


def mutates(f):
    '''Decorator for functions that mutate :class:`StringCounter`.

    This raises :exc:`~dossier.fc.exceptions.ReadOnlyException` if
    the object is read-only, and increments the generation counter
    otherwise.
    '''
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        if self.read_only:
            raise ReadOnlyException()
        self.next_generation()
        return f(self, *args, **kwargs)
    return wrapper


class StringCounter(Counter):
    '''Simple counter based on exact string matching.

    This is a subclass of :class:`collections.Counter` that includes a
    generation counter so that it can be used in a cache.

    :class:`StringCounter` is the default feature type in a feature
    collection, so you typically don't have to instantiate a
    :class:`StringCounter` explicitly::

        fc = FeatureCollection()
        fc['NAME']['John Smith'] += 1

    But instantiating directly works too::

        sc = StringCounter()
        sc['John Smith'] += 1

        fc = FeatureCollection({'NAME': sc})
        fc['NAME']['John Smith'] += 1
        assert fc['NAME']['John Smith'] == 2

    Note that instances of this class support all the methods defined
    for a :class:`collections.Counter`, but only the ones unique to
    :class:`StringCounter` are listed here.

    .. automethod:: __init__
    .. automethod:: truncate_most_common

    .. attribute:: read_only

        Flag indicating whether this collection is read-only.

        This flag always begins as :const:`False`, it cannot be set
        via the constructor for compatibility with
        :class:`collections.Counter`.  If this flag is set, then any
        operations that mutate it will raise
        :exc:`~dossier.fc.exceptions.ReadOnlyException`.

    .. attribute:: generation

        Generation number for this counter instance.

        This number is incremented by every operation that
        mutates the counter object.  If two collections are the
        same object and have the same generation number, then
        they are identical.

        Having this property allows a pair of `id(sc)` and the
        generation to be an immutable hashable key for things like
        memoization operations, accounting for the possibility of the
        counter changing over time.

        >>> sc = StringCounter({'a': 1})
        >>> cache = {(id(sc), sc.generation): 1}
        >>> (id(sc), sc.generation) in cache
        True
        >>> sc['a']
        1
        >>> (id(sc), sc.generation) in cache
        True
        >>> sc['a'] += 1
        >>> sc['a']
        2
        >>> (id(sc), sc.generation) in cache
        False

    '''

    current_generation = 0
    '''Class-static generation number.

    Each mutation of a StringCounter increments this generation,
    and sets the counter's current generation to this value.
    See :meth:`next_generation` for details.
    '''

    def __init__(self, *args, **kwargs):
        '''Initialize a :class:`StringCounter` with existing counts::

            >>> sc = StringCounter(a=4, b=2, c=0)
            >>> sc['b']
            2

        See the documentation for :class:`collections.Counter` for more
        examples.
        '''

        self.read_only = False
        self.generation = self.current_generation
        super(StringCounter, self).__init__(*args, **kwargs)

    def next_generation(self):
        '''Increment the generation counter on this collection.'''
        self.current_generation += 1
        self.generation = self.current_generation

    @mutates
    def __delitem__(self, key):
        return super(StringCounter, self).__delitem__(key)

    @staticmethod
    def _fix_key(key):
        '''Normalize keys to Unicode strings.'''
        if isinstance(key, unicode):
            return key
        if isinstance(key, str):
            # On my system, the default encoding is `ascii`, so let's
            # explicitly say UTF-8?
            return unicode(key, 'utf-8')
        raise TypeError(key)

    @mutates
    def __setitem__(self, key, value):
        key = self._fix_key(key)
        return super(StringCounter, self).__setitem__(key, value)

    @mutates
    def pop(self, key):
        return super(StringCounter, self).pop(key)

    @mutates
    def popitem(self, key):
        return super(StringCounter, self).popitem(key)

    @mutates
    def subtract(self, other):
        return super(StringCounter, self).subtract(other)

    @mutates
    def update(self, iterable=None, **kwargs):
        # Force all keys into Unicode strings before calling base
        # class implementation; if kwargs is non-empty then the base
        # class will call this method again
        if iterable:
            if isinstance(iterable, Mapping):
                new_iterable = {}
                for (k, v) in iterable.iteritems():
                    new_iterable[self._fix_key(k)] = v
                iterable = new_iterable
            else:
                iterable = (self._fix_key(k) for k in iterable)
        return super(StringCounter, self).update(iterable, **kwargs)

    def __add__(self, other):
        result = super(StringCounter, self).__add__(other)
        return StringCounter(result)

    def __sub__(self, other):
        result = super(StringCounter, self).__sub__(other)
        return StringCounter(result)

    @mutates
    def __imul__(self, coef):
        for k in self.keys():
            self[k] *= coef
        return self

    @mutates
    def truncate_most_common(self, truncation_length):
        '''
        Sorts the counter and keeps only the most common items up to
        ``truncation_length`` in place.

        :type truncation_length: int
        '''
        keep_keys = set(v[0] for v in self.most_common(truncation_length))
        for key in self.keys():
            if key not in keep_keys:
                self.pop(key)


class StringCounterSerializer(object):
    def __init__(self):
        raise NotImplementedError()

    loads = StringCounter

    @staticmethod
    def dumps(sc):
        return dict(sc)

    constructor = StringCounter


class NestedStringCounter(MutableMapping):
    '''A mapping from string to string counter.
    '''
    def __init__(self, data=None):
        self.data = {}
        if data is not None:
            for fname, counter in data.items():
                if fname not in self.data:
                    self.data[fname] = StringCounter()
                for key, count in counter.iteritems():
                    self.data[fname][key] = count

    def to_nested_dict(self):
        dumped = {}
        for key, counter in self.data.iteritems():
            dumped[key] = dict(counter)
        return dumped

    @staticmethod
    def from_nested_dict(d):
        return NestedStringCounter(data=d)

    def __getitem__(self, key):
        return self.data.get(uni(key)) or self.__missing__(key)

    def __missing__(self, key):
        v = StringCounter()
        self[uni(key)] = v
        return v

    def __contains__(self, key):
        return key in self.data

    def __setitem__(self, key, value):
        self.data[uni(key)] = value

    def __delitem__(self, key):
        del self.data[uni(key)]

    def __len__(self): return len(self.data)

    def __iter__(self): return iter(self.data)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, repr(self.data))

    def __add__(self, other):
        new_nsc = NestedStringCounter(self.to_nested_dict())
        for key, counter in other.items():
            if key in new_nsc:
                new_nsc[key] += counter
            else:
                new_nsc[key] = counter
        return new_nsc

class NestedStringCounterSerializer(object):
    '''Serialize nested string counters.'''

    def __init__(self):
        raise NotImplementedError()

    dumps = NestedStringCounter.to_nested_dict
    constructor = NestedStringCounter

    @staticmethod
    def loads(d):
        return NestedStringCounter.from_nested_dict(d)



'''tests for NestedStringCounter

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

'''

from dossier.fc import FeatureCollection
from dossier.fc.string_counter import NestedStringCounter, StringCounter, \
    NestedStringCounterSerializer

def test_nested_fcdefault():
    fc = FeatureCollection()
    assert isinstance(fc['%names'], NestedStringCounter)

def test_nsc_roundtrip():
    fc = FeatureCollection()
    fc['#testing'] = NestedStringCounter()
    fc['#testing']['foo'] = StringCounter({'foo': 1})
    fc['#testing']['bar'] = StringCounter({'foo': 2, 'bar': 1})
    dumped = fc.dumps()
    assert FeatureCollection.loads(dumped) == fc

def test_nsc():
    data = {'foo': StringCounter({'bar': 1}),
            'baz': StringCounter({'bar': 1})}

    nsc = NestedStringCounter(data=data)
    out = NestedStringCounterSerializer.dumps(nsc)
    nsc2 = NestedStringCounterSerializer.loads(out)

    assert nsc2 is not nsc
    assert nsc2 == nsc

def test_nsc_add():
    data1 = {'foo': StringCounter({'bar': 1}),
             'baz': StringCounter({'bar': 1})}
    data2 = {'foo': StringCounter({'foo': 1, 'bar': 2}),
             'bar': StringCounter({'baz': 1})}

    nsc1 = NestedStringCounter(data=data1)
    nsc2 = NestedStringCounter(data=data2)

    # Sum of two counters.
    added = nsc1 + nsc2
    assert added['foo'] == StringCounter({'bar': 3, 'foo': 1})
    assert added['baz'] == StringCounter({'bar': 1})
    assert added['bar'] == StringCounter({'baz': 1})

    # Original counters unchanged.
    assert nsc1['foo'] == StringCounter({'bar': 1})
    assert nsc1['baz'] == StringCounter({'bar': 1})
    assert 'bar' not in nsc1

    assert nsc2['foo'] == StringCounter({'foo': 1, 'bar': 2})
    assert nsc2['bar'] == StringCounter({'baz': 1})
    assert 'baz' not in nsc2

    # Test in-place.
    nsc1 += nsc2
    assert nsc1 == added



'''GeoCoords carries (longitude, latitude) coordinates

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.
'''
from __future__ import absolute_import, division, print_function

from collections import MutableMapping
from itertools import ifilter, imap

from dossier.fc.exceptions import uni

class GeoCoords(MutableMapping):
    '''This maps string values to lists of space-time coordinates:

        (longitude, latitude, altitude, time)

    where the spatial coordinates are expected to be Earth coordinates
    in WGS84.  `None` values *are allowed* in the four-tuples, so, for
    example, if your data lacks altitude or time, set those values to
    `None`.

    '''
    allowed_types = (float, int, long, type(None))

    def __init__(self, data=None):
        self.data = {}
        if data is not None:
            for fname, coords in data.items():
                if fname not in self.data:
                    self.data[fname] = []
                for coord in coords:
                    if not len(coord) == 4 or \
                       not all(map(lambda x: isinstance(x, self.allowed_types), 
                                   coord)):
                    # could add range checking here
                        raise Exception('expecting four-tuples of lon, '
                                        'lat, alt, time and got: %s' % repr(coord))
                    self.data[fname].append(tuple(coord))

    def to_dict(self):
        return self.data

    @staticmethod
    def from_dict(data):
        return GeoCoords(data)

    # Methods for satisfying `MutableMapping`.

    def __getitem__(self, k):
        return self.data.get(uni(k)) or self.__missing__(k)

    def __missing__(self, k):
        v = []
        self[uni(k)] = v
        return v

    def __setitem__(self, k, v):
        self.data[uni(k)] = v

    def __delitem__(self, k): 
        del self.data[uni(k)]

    def __len__(self): return len(self.data)
    def __iter__(self): return iter(self.data)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, repr(self.data))


class GeoCoordsSerializer(object):
    '''Serialization for geocoords.'''
    def __init__(self):
        raise NotImplementedError()

    dumps = GeoCoords.to_dict
    constructor = GeoCoords

    @staticmethod
    def loads(d):
        return GeoCoords.from_dict(d)



'''dossier.fc Feature Collections

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

tests to verify that StringCounter.generation changes when it should
'''
from __future__ import absolute_import, division, print_function

import pytest

from dossier.fc.tests import counter_type


@pytest.fixture
def setup(counter_type):
    c1 = counter_type({'dog': 5, 'cat': 3})
    c2 = counter_type({'dog': 8, 'fish': 2})
    return (c1, c2)


def has_gen(obj):
    return hasattr(obj, 'generation')


def test_generation_iadd(setup):
    c1, c2 = setup
    if not has_gen(c1) or not has_gen(c2):
        return

    h11 = (id(c1), c1.generation)
    h21 = (id(c2), c2.generation)
    c1 += c2
    h12 = (id(c1), c1.generation)
    h22 = (id(c2), c2.generation)

    assert h11 != h12
    assert h21 == h22, c2


def test_generation_isub(setup):
    c1, c2 = setup
    if not has_gen(c1) or not has_gen(c2):
        return

    h11 = (id(c1), c1.generation)
    h21 = (id(c2), c2.generation)
    c1 -= c2
    h12 = (id(c1), c1.generation)
    h22 = (id(c2), c2.generation)

    assert h11 != h12
    assert h21 == h22, c2


def test_generation_add(setup):
    c1, c2 = setup
    if not has_gen(c1) or not has_gen(c2):
        return

    h11 = (id(c1), c1.generation)
    h21 = (id(c2), c2.generation)
    c3 = c1 + c2
    h12 = (id(c1), c1.generation)
    h22 = (id(c2), c2.generation)
    h32 = (id(c3), c3.generation)

    assert h11 == h12
    assert h11 != h32
    assert h21 == h22
    assert h21 != h32


def test_generation_pop(counter_type):
    c1 = counter_type(dict(cat=5, dog=3))
    if not has_gen(c1):
        return

    h11 = (id(c1), c1.generation)
    c1.pop('cat')
    h12 = (id(c1), c1.generation)
    assert h11 != h12, c1


def test_generation_missing(counter_type):
    c1 = counter_type(dict(cat=5, dog=3))
    if not has_gen(c1):
        return

    assert not ('bird' in c1)
    h11 = (id(c1), c1.generation)
    assert c1['bird'] == 0
    c1['bird'] += 1
    h12 = (id(c1), c1.generation)
    assert h11 != h12, c1


def test_generation_delitem(counter_type):
    c1 = counter_type(dict(cat=5, dog=3))
    if not has_gen(c1):
        return

    assert not ('bird' in c1)
    h11 = (id(c1), c1.generation)
    del c1['cat']
    h12 = (id(c1), c1.generation)
    assert h11 != h12, c1



#!/usr/bin/env python

import os

from setuptools import setup, find_packages

from version import get_git_version
VERSION, SOURCE_LABEL = get_git_version()
PROJECT = 'dossier.fc'
AUTHOR = 'Diffeo, Inc.'
AUTHOR_EMAIL = 'support@diffeo.com'
URL = 'http://github.com/dossier/dossier.fc'
DESC = 'Feature collections for DossierStack'


def read_file(file_name):
    file_path = os.path.join(
        os.path.dirname(__file__),
        file_name
        )
    return open(file_path).read()

setup(
    name=PROJECT,
    version=VERSION,
    description=DESC,
    license='MIT/X11 license http://opensource.org/licenses/MIT',
    long_description=read_file('README.md'),
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages(),
    namespace_packages=['dossier'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
    ],
    install_requires=[
        'cbor >= 0.1.13',
        'streamcorpus >= 0.3.27',
    ],
    extras_require={
        'unittest': [
            'pytest',
        ],
        ':python_version=="2.6"': [
            'backport_collections',
        ],
    },
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'dossier.fc = dossier.fc.dump:main',
            'dossier.fc.test = dossier.fc.tests.run:main',
        ],
    },
)



from __future__ import absolute_import, division, print_function

from dossier.fc import FeatureCollection, FeatureTokens
import streamcorpus


def test_ft_fcdefault():
    fc = FeatureCollection()
    assert isinstance(fc['@NAME'], FeatureTokens)


def test_ft_default():
    fo = FeatureTokens()
    assert fo['foo'] == []


def test_ft_roundtrip():
    fc = FeatureCollection()
    fc['@NAME']['foo'].append([
        ('nltk', 5, 2),
    ])
    fc2 = FeatureCollection.loads(fc.dumps())
    assert fc['@NAME'] == fc2['@NAME']


def test_ft_with_stream_item():
    si = streamcorpus.make_stream_item('2005-01-01T05:06:07.0Z', 'abc')
    tokens = [
        streamcorpus.Token(offsets={
            streamcorpus.OffsetType.XPATH_CHARS: streamcorpus.Offset(
                type=streamcorpus.OffsetType.XPATH_CHARS,
                xpath='/html[1]/body[1]/p[1]/b[1]/text()[1]',
                first=0,
                xpath_end='/html[1]/body[1]/p[1]/text()[1]',
                xpath_end_offset=2),
        }),
        streamcorpus.Token(offsets={
            streamcorpus.OffsetType.XPATH_CHARS: streamcorpus.Offset(
                type=streamcorpus.OffsetType.XPATH_CHARS,
                xpath='/html[1]/body[1]/p[1]/b[2]/text()[1]',
                first=0,
                xpath_end='/html[1]/body[1]/p[1]/text()[2]',
                xpath_end_offset=4),
        }),
    ]
    si.body.sentences = {'test': [streamcorpus.Sentence(tokens=tokens)]}
    si.body.clean_html = '<html><body><p><b>T</b>om ' \
                         '<b>B</b>rady</p></body></html>'

    ft = FeatureTokens()
    ft['tom brady'].append([('test', 0, 0), ('test', 0, 1)])
    assert next(ft.xpath_slices(si, 'tom brady')) == 'Tom Brady'
